# -*- coding: utf-8 -*-

"""
Unit test for the Build models
"""

import os
import sys

sys.path.append(os.path.abspath(os.path.dirname(__file__)))

from test_api import InvenTreeTestCase  # noqa: E402

from inventree.base import Attachment  # noqa: E402
from inventree.build import Build  # noqa: E402


class BuildOrderTest(InvenTreeTestCase):
    """
    Unit tests for Build model
    """

    def get_build(self):
        """
        Return a BuildOrder from the database
        If a build does not already exist, create a new one
        """

        builds = Build.list(self.api)

        n = len(builds)

        if n == 0:
            # Create a new build order
            build = Build.create(
                self.api,
                {
                    "title": "Automated test build",
                    "part": 25,
                    "quantity": 100,
                    "reference": f"BO-{n+1:04d}",
                }
            )
        else:
            build = builds[-1]

        return build

    def test_list_builds(self):

        build = self.get_build()

        self.assertIsNotNone(build)

        builds = Build.list(self.api)

        self.assertGreater(len(builds), 0)

    def test_build_attachment(self):
        """
        Test that we can upload an attachment against a Build
        """

        if self.api.api_version < Attachment.MIN_API_VERSION:
            return

        build = self.get_build()

        n = len(build.getAttachments())

        # Upload *this* file
        fn = os.path.join(os.path.dirname(__file__), 'test_build.py')

        response = build.uploadAttachment(fn, comment='A self referencing upload!')

        self.assertEqual(response['model_type'], 'build')
        self.assertEqual(response['model_id'], build.pk)
        self.assertEqual(response['comment'], 'A self referencing upload!')

        self.assertEqual(len(build.getAttachments()), n + 1)

    def test_build_cancel(self):
        """
        Test cancelling a build order.
        """

        n = len(Build.list(self.api))

        # Create a new build order
        build = Build.create(
            self.api,
            {
                "title": "Automated test build",
                "part": 25,
                "quantity": 100,
                "reference": f"BO-{n+1:04d}"
            }
        )

        # Cancel
        build.cancel()

        # Check status
        self.assertEqual(build.status, 30)
        self.assertEqual(build.status_text, 'Cancelled')

    def test_build_complete(self):
        """
        Test completing a build order.
        """

        n = len(Build.list(self.api))

        # Create a new build order
        build = Build.create(
            self.api,
            {
                "title": "Automated test build",
                "part": 25,
                "quantity": 100,
                "reference": f"BO-{n+1:04d}"
            }
        )

        # Check that build status is pending
        self.assertEqual(build.status, 10)

        if self.api.api_version >= 233:
            # Issue the build order
            build.issue()
            self.assertEqual(build.status, 20)

            # Mark build order as "on hold" again
            build.hold()
            self.assertEqual(build.status, 25)

            # Issue again
            build.issue()
            self.assertEqual(build.status, 20)

        # Complete the build, even though it is not completed
        build.complete(accept_unallocated=True, accept_incomplete=True)

        # Check status
        self.assertEqual(build.status, 40)
        self.assertEqual(build.status_text, 'Complete')



# -*- coding: utf-8 -*-

import os
import sys

import requests
from requests.exceptions import HTTPError

try:
    import Image
except ImportError:
    from PIL import Image

sys.path.append(os.path.abspath(os.path.dirname(__file__)))

from test_api import InvenTreeTestCase  # noqa: E402

from inventree.base import Attachment  # noqa: E402
from inventree.company import SupplierPart  # noqa: E402
from inventree.part import InternalPrice  # noqa: E402
from inventree.part import (BomItem, Parameter,  # noqa: E402
                            ParameterTemplate, Part,
                            PartCategory, PartCategoryParameterTemplate,
                            PartRelated, PartTestTemplate)
from inventree.stock import StockItem  # noqa: E402


class PartCategoryTest(InvenTreeTestCase):
    """Tests for PartCategory models"""

    def test_part_cats(self):
        """
        Tests for category filtering
        """

        # All categories
        cats = PartCategory.list(self.api)
        n = len(cats)
        self.assertTrue(len(cats) >= 9)

        # Check that the 'count' method returns the same result
        self.assertEqual(n, PartCategory.count(self.api))

        # Filtered categories must be fewer than *all* categories
        cats = PartCategory.list(self.api, parent=1)

        self.assertGreater(len(cats), 0)
        self.assertLess(len(cats), n)

    def test_elec(self):
        electronics = PartCategory(self.api, 1)

        # This is a top-level category, should not have a parent!
        self.assertIsNone(electronics.getParentCategory())
        self.assertEqual(electronics.name, "Electronics")

        children = electronics.getChildCategories()
        self.assertGreaterEqual(len(children), 3)

        for child in children:
            self.assertEqual(child.parent, 1)

        child = PartCategory(self.api, pk=3)
        self.assertEqual(child.name, 'Capacitors')
        self.assertEqual(child.getParentCategory().pk, electronics.pk)

        # Grab all child categories
        children = PartCategory.list(self.api, parent=child.pk)

        n = len(children)

        for c in children:
            self.assertEqual(c.parent, child.pk)

        # Create some new categories under this one
        for idx in range(3):
            name = f"Subcategory {n+idx}"

            cat = PartCategory.create(self.api, {
                "parent": child.pk,
                "name": name,
                "description": "A new subcategory",
            })

            self.assertEqual(cat.parent, child.pk)
            self.assertEqual(cat.name, name)

            # Edit the name of the new location
            cat.save({
                "name": f"{name}_suffix",
            })

            # Reload from server, and check
            cat.reload()
            self.assertEqual(cat.name, f"{name}_suffix")

        # Number of children should have increased!
        self.assertEqual(len(child.getChildCategories()), n + 3)

    def test_caps(self):

        cat = PartCategory(self.api, 6)
        self.assertEqual(cat.name, "Transceivers")
        parts = cat.getParts()

        n_parts = len(parts)

        for p in parts:
            self.assertEqual(p.category, cat.pk)

        # Create some new parts
        for i in range(10):

            name = f"Part_{cat.pk}_{n_parts + i}"

            prt = Part.create(self.api, {
                "category": cat.pk,
                "name": name,
                "description": "A new part in this category",
            })

            self.assertIsNotNone(prt)

            self.assertEqual(prt.name, name)

        parts = cat.getParts()

        self.assertEqual(len(parts), n_parts + 10)

    def test_part_category_parameter_templates(self):
        """Unit tests for the PartCategoryParameterTemplate model"""

        electronics = PartCategory(self.api, pk=3)

        # Ensure there are some parameter templates associated with this category
        templates = electronics.getCategoryParameterTemplates(fetch_parent=False)

        if len(templates) == 0:
            for name in ['wodth', 'lungth', 'herght']:
                template = ParameterTemplate.create(self.api, data={
                    'name': name,
                    'units': 'mm',
                })

                pcpt = PartCategoryParameterTemplate.create(
                    self.api,
                    data={
                        'category': electronics.pk,
                        'parameter_template': template.pk,
                        'default_value': 123,
                    }
                )

                # Check that model lookup functions work
                self.assertEqual(pcpt.getCategory().pk, electronics.pk)
                self.assertEqual(pcpt.getTemplate().pk, template.pk)

            # Reload
            templates = electronics.getCategoryParameterTemplates(fetch_parent=False)

        self.assertTrue(len(templates) >= 3)

        # Check child categories
        children = electronics.getChildCategories()

        self.assertTrue(len(children) > 0)

        for child in children:
            child_templates = child.getCategoryParameterTemplates(fetch_parent=True)
            self.assertTrue(len(child_templates) >= 3)


class PartTest(InvenTreeTestCase):
    """Tests for Part models"""

    def test_part_get_functions(self):
        """Test various functions of Part class, mostly starting with get...
        These are wrappers for other functions, so the testing of details of the function should
        be done elsewhere."""

        # Get list of parts
        parts = Part.list(self.api)

        functions = {
            'getSupplierParts': SupplierPart,
            'getBomItems': BomItem,
            'isUsedIn': BomItem,
            'getStockItems': StockItem,
            'getParameters': Parameter,
            'getRelated': PartRelated,
            'getInternalPriceList': InternalPrice,
        }

        if self.api.api_version >= Attachment.MIN_API_VERSION:
            functions['getAttachments'] = Attachment

        # For each part in list, test some functions
        for p in parts:

            for fnc, res in functions.items():
                A = getattr(p, fnc)()
                # Make sure a list is returned
                self.assertIsInstance(A, list)
                for a in A:
                    # Make sure any result is of the right class
                    self.assertIsInstance(a, res)

    def test_access_erors(self):
        """
        Test that errors are flagged when we try to access an invalid part
        """

        with self.assertRaises(TypeError):
            Part(self.api, 'hello')

        with self.assertRaises(ValueError):
            Part(self.api, -1)

        # Try to access a Part which does not exist
        with self.assertRaises(requests.exceptions.HTTPError):
            Part(self.api, 9999999999999)

    def test_fields(self):
        """
        Test field names via OPTIONS request
        """

        field_names = Part.fieldNames(self.api)

        self.assertIn('active', field_names)
        self.assertIn('revision', field_names)
        self.assertIn('full_name', field_names)
        self.assertIn('IPN', field_names)

    def test_options(self):
        """Extends tests for OPTIONS model metadata"""

        # Check for field which does not exist
        with self.assertLogs():
            Part.fieldInfo('abcde', self.api)

        active = Part.fieldInfo('active', self.api)

        self.assertEqual(active['type'], 'boolean')
        self.assertEqual(active['label'], 'Active')
        self.assertEqual(active['default'], True)

        for field_name in [
            'name',
            'description',
            'component',
            'assembly',
        ]:
            field = Part.fieldInfo(field_name, self.api)

            # Check required field attributes
            for attr in ['type', 'required', 'read_only', 'label', 'help_text']:
                self.assertIn(attr, field)

    def test_pagination(self):
        """ Test that we can paginate the queryset by specifying a 'limit' parameter"""

        parts = Part.list(self.api, limit=5)
        self.assertEqual(len(parts), 5)

        for p in parts:
            self.assertTrue(type(p) is Part)

    def test_part_list(self):
        """
        Check that we can list Part objects,
        and apply certain filters
        """

        parts = Part.list(self.api)
        self.assertTrue(len(parts) >= 19)

        parts = Part.list(self.api, category=5)

        n = len(parts)

        for i in range(5):
            prt = Part.create(self.api, {
                "category": 5,
                "name": f"Special Part {n+i}",
                "description": "A new part in this category!",
            })

            self.assertEqual(prt.category, 5)
            cat = prt.getCategory()
            self.assertEqual(cat.pk, 5)

            parts = cat.getParts()

            self.assertGreaterEqual(len(parts), i + 1)

    def test_part_edit(self):
        """
        Test that we can edit a part
        """

        # Select a part
        p = Part.list(self.api)[-1]

        name = p.name

        # Adjust the name
        if len(name) < 40:
            name += '_append'
        else:
            name = name[:-10]

        p.save(
            data={
                'name': name,
                'description': 'A new description'
            },
        )
        p.reload()

        self.assertEqual(p.name, name)
        self.assertEqual(p.description, 'A new description')

    def test_default_values(self):
        """
        Test that the DRF framework will correctly insert the default values
        """

        n = len(Part.list(self.api))

        # Create a part without specifying 'active' and 'virtual' fields
        p = Part.create(
            self.api,
            {
                'name': f"Part_{n}_default_test",
                'category': 1,
                'description': "Some part thingy",
            }
        )

        self.assertEqual(p.active, True)
        self.assertEqual(p.virtual, False)

        # Set both to false
        p = Part.create(
            self.api,
            {
                'name': f"Part_{n}_default_test_2",
                'category': 1,
                'description': 'Setting fields to false',
                'active': False,
                'virtual': False,
            }
        )

        self.assertFalse(p.active)
        self.assertFalse(p.virtual)

        # Set both to true
        p = Part.create(
            self.api,
            {
                'name': f"Part_{n}_default_test_3",
                'category': 1,
                'description': 'Setting fields to true',
                'active': True,
                'virtual': True,
            }
        )

        self.assertTrue(p.active)
        self.assertTrue(p.virtual)

    def test_part_delete(self):
        """
        Test we can create and delete a Part instance via the API
        """

        n = len(Part.list(self.api))

        # Create a new part
        # We do not specify 'active' value so it will default to True
        p = Part.create(
            self.api,
            {
                'name': 'Delete Me',
                'description': 'Not long for this world!',
                'category': 1,
            }
        )

        self.assertIsNotNone(p)
        self.assertIsNotNone(p.pk)

        self.assertEqual(len(Part.list(self.api)), n + 1)

        # Cannot delete - part is 'active'!
        with self.assertRaises(requests.exceptions.HTTPError) as ar:
            response = p.delete()

        self.assertIn("Cannot delete this part as it is still active", str(ar.exception))

        p.save(data={'active': False})
        response = p.delete()
        self.assertEqual(response.status_code, 204)

        # And check that the part has indeed been deleted
        self.assertEqual(len(Part.list(self.api)), n)

    def test_image_upload(self):
        """
        Test image upload functionality for Part model
        """

        # Grab the first part
        p = Part.list(self.api)[0]

        # Ensure the part does *not* have an image associated with it
        p.save(data={'image': None})

        # Create a dummy file (not an image)
        with open('dummy_image.jpg', 'w') as dummy_file:
            dummy_file.write("hello world")

        # Attempt to upload an image
        with self.assertRaises(requests.exceptions.HTTPError):
            response = p.uploadImage("dummy_image.jpg")

        # Now, let's actually upload a real image
        img = Image.new('RGB', (128, 128), color='red')
        img.save('dummy_image.png')

        response = p.uploadImage("dummy_image.png")

        self.assertIsNotNone(response)
        self.assertIsNotNone(p['image'])
        self.assertIn('dummy_image', p['image'])

        # Re-download the image file
        fout = 'test/output.png'

        if os.path.exists(fout):
            # Delete the file if it already exists
            os.remove(fout)

        response = p.downloadImage(fout)
        self.assertTrue(response)

        self.assertTrue(os.path.exists(fout))

        # Attempt to re-download
        with self.assertRaises(FileExistsError):
            p.downloadImage(fout)

        # Download, with overwrite enabled
        p.downloadImage(fout, overwrite=True)

    def test_part_attachment(self):
        """
        Check that we can upload attachment files against the part
        """

        if self.api.api_version < Attachment.MIN_API_VERSION:
            return

        prt = Part(self.api, pk=1)
        attachments = prt.getAttachments()

        for a in attachments:
            self.assertEqual(a.model_type, 'part')
            self.assertEqual(a.model_id, prt.pk)

        n = len(attachments)

        # Test that attempting to upload an invalid file fails
        with self.assertRaises(FileNotFoundError):
            prt.uploadAttachment('test-file.txt')

        # Check that no new files have been uploaded
        self.assertEqual(len(prt.getAttachments()), n)

        # Test that we can upload a file by filename, directly from the Part instance
        filename = os.path.join(os.path.dirname(__file__), 'docker-compose.yml')

        response = prt.uploadAttachment(
            filename,
            comment='Uploading a file'
        )

        self.assertIsNotNone(response)

        pk = response['pk']

        # Check that a new attachment has been created!
        attachment = Attachment(self.api, pk=pk)
        self.assertTrue(attachment.is_valid())

        # Download the attachment to a local file!
        dst = os.path.join(os.path.dirname(__file__), 'test.tmp')
        attachment.download(dst, overwrite=True)

        self.assertTrue(os.path.exists(dst))
        self.assertTrue(os.path.isfile(dst))

        with self.assertRaises(FileExistsError):
            # Attempt to download the file again, but without overwrite option
            attachment.download(dst)

    def test_part_link_attachment(self):
        """
        Check that we can add an external link attachment to the part
        """

        if self.api.api_version < Attachment.MIN_API_VERSION:
            return

        test_link = "https://inventree.org/"
        test_comment = "inventree.org"

        # Add valid external link attachment
        part = Part(self.api, pk=1)
        response = part.addLinkAttachment(test_link, comment=test_comment)
        self.assertIsNotNone(response)

        # Check that the attachment has been created
        attachment = Attachment(self.api, pk=response["pk"])
        self.assertTrue(attachment.is_valid())
        self.assertEqual(attachment.link, test_link)
        self.assertEqual(attachment.comment, test_comment)

    def test_set_price(self):
        """
        Tests that an internal price can be set for a part
        """

        test_price = 100.0
        test_quantity = 1

        # Grab the first part
        p = Part.list(self.api)[0]

        # Grab all internal prices for the part
        ip = InternalPrice.list(self.api, part=p.pk)

        # Delete any existing prices
        for price in ip:
            self.assertEqual(type(price), InternalPrice)
            price.delete()

        # Ensure that no part has an internal price
        ip = InternalPrice.list(self.api, part=p.pk)
        self.assertEqual(len(ip), 0)

        # Set the internal price
        p.setInternalPrice(test_quantity, test_price)

        # Ensure that the part has an internal price
        ip = InternalPrice.list(self.api, part=p.pk)
        self.assertEqual(len(ip), 1)

        # Grab the internal price
        ip = ip[0]

        self.assertEqual(ip.quantity, test_quantity)
        self.assertEqual(ip.part, p.pk)
        ip_price_clean = float(ip.price)
        self.assertEqual(ip_price_clean, test_price)

    def test_parameters(self):
        """
        Test setting and getting Part parameter templates, as well as parameter values
        """

        # Count number of existing Parameter Templates
        existingTemplates = len(ParameterTemplate.list(self.api))

        # Create new parameter template - this will fail, no name given
        with self.assertRaises(HTTPError):
            parametertemplate = ParameterTemplate.create(self.api, data={'units': "kg A"})

        # Now create a proper parameter template
        parametertemplate = ParameterTemplate.create(self.api, data={'name': f'Test parameter no {existingTemplates}', 'units': "kg A"})

        # result should not be None
        self.assertIsNotNone(parametertemplate)

        # Count should be one higher now
        self.assertEqual(len(ParameterTemplate.list(self.api)), existingTemplates + 1)

        # Grab the first part
        p = Part.list(self.api)[0]

        # Count number of parameters
        existingParameters = len(p.getParameters())

        # Define parameter value for this part - without all required values
        with self.assertRaises(HTTPError):
            Parameter.create(self.api, data={'part': p.pk, 'template': parametertemplate.pk})

        # Define parameter value for this part - without all required values
        with self.assertRaises(HTTPError):
            Parameter.create(self.api, data={'part': p.pk, 'data': 10})

        # Define w. required values - integer
        param = Parameter.create(self.api, data={'part': p.pk, 'template': parametertemplate.pk, 'data': 10})

        # Unit should be equal
        self.assertEqual(param.getunits(), 'kg A')

        # result should not be None
        self.assertIsNotNone(param)

        # Same parameter for same part - should fail
        # Define w. required values - string
        with self.assertRaises(HTTPError):
            Parameter.create(self.api, data={'part': p.pk, 'template': parametertemplate.pk, 'data': 'String value'})

        # Number of parameters should be one higher than before
        self.assertEqual(len(p.getParameters()), existingParameters + 1)

        # Delete the parameter
        param.delete()

        # Check count
        self.assertEqual(len(p.getParameters()), existingParameters)

        # Delete the parameter template
        parametertemplate.delete()

        # Check count
        self.assertEqual(len(ParameterTemplate.list(self.api)), existingTemplates)

    def test_metadata(self):
        """Test Part instance metadata"""

        # Grab the first available part
        part = Part.list(self.api, limit=1)[0]

        part.setMetadata(
            {
                "foo": "bar",
            },
            overwrite=True,
        )

        metadata = part.getMetadata()

        # Check that the metadata has been overwritten
        self.assertEqual(len(metadata.keys()), 1)

        self.assertEqual(metadata['foo'], 'bar')

        # Now 'patch' in some metadata
        part.setMetadata(
            {
                'hello': 'world',
            },
        )

        part.setMetadata(
            {
                'foo': 'rab',
            }
        )

        metadata = part.getMetadata()

        self.assertEqual(len(metadata.keys()), 2)
        self.assertEqual(metadata['foo'], 'rab')
        self.assertEqual(metadata['hello'], 'world')

    def test_part_related(self):
        """Test add related function"""

        parts = Part.list(self.api)

        # First, ensure *all* related parts are deleted
        for relation in PartRelated.list(self.api):
            relation.delete()

        # Take two parts, make them related
        # Try with pk values
        ret = PartRelated.add_related(self.api, parts[0].pk, parts[1].pk)
        self.assertTrue(ret)

        # Take two parts, make them related
        # Try with Part object
        ret = PartRelated.add_related(self.api, parts[2], parts[3])
        self.assertTrue(ret)

        # Take the same part twice, should fail
        with self.assertRaises(HTTPError):
            ret = PartRelated.add_related(self.api, parts[3], parts[3])

    def test_get_requirements(self):
        """Test getRequirements function for parts"""

        # Get first part
        prt = Part.list(self.api, limit=1)[0]

        # Get requirements list
        req = prt.getRequirements()

        # Check for expected content
        self.assertIsInstance(req, dict)
        self.assertIn('available_stock', req)
        self.assertIn('on_order', req)
        self.assertIn('required_build_order_quantity', req)
        self.assertIn('allocated_build_order_quantity', req)
        self.assertIn('required_sales_order_quantity', req)
        self.assertIn('allocated_sales_order_quantity', req)
        self.assertIn('allocated', req)
        self.assertIn('required', req)


class PartBarcodeTest(InvenTreeTestCase):
    """Tests for Part barcode functionality"""

    def test_barcode_assign(self):
        """Tests for assigning barcodes to Part instances"""

        barcode = 'ABCD-1234-XYZ'

        # Grab a part from the database
        part_1 = Part(self.api, pk=1)

        # First ensure that there is *no* barcode assigned to this item
        part_1.unassignBarcode()

        # Assign a barcode to this part (should auto-reload)
        response = part_1.assignBarcode(barcode)

        self.assertEqual(response['success'], 'Assigned barcode to part instance')
        self.assertEqual(response['barcode_data'], barcode)

        # Attempt to assign the same barcode to a different part (should error)
        part_2 = Part(self.api, pk=2)

        # Ensure this part does not have an associated barcode
        part_2.unassignBarcode()

        with self.assertRaises(HTTPError):
            response = part_2.assignBarcode(barcode)

        # Scan the barcode (should point back to part_1)
        response = self.api.scanBarcode(barcode)

        self.assertEqual(response['barcode_data'], barcode)
        self.assertEqual(response['part']['pk'], 1)

        # Unassign from part_1
        part_1.unassignBarcode()

        # Now assign to part_2
        response = part_2.assignBarcode(barcode)
        self.assertEqual(response['barcode_data'], barcode)

        # Scan again
        response = self.api.scanBarcode(barcode)
        self.assertEqual(response['part']['pk'], 2)

        # Unassign from part_2
        part_2.unassignBarcode()

        # Scanning this time should yield no results
        with self.assertRaises(HTTPError):
            response = self.api.scanBarcode(barcode)


class PartTestTemplateTest(InvenTreeTestCase):
    """Tests for PartTestTemplate functionality"""

    def test_generateKey(self):
        """Tests for generating a key for a PartTestTemplate"""

        self.assertEqual(PartTestTemplate.generateTestKey('bob'), 'bob')
        self.assertEqual(PartTestTemplate.generateTestKey('bob%35'), 'bob35')
        self.assertEqual(PartTestTemplate.generateTestKey('bo b%35'), 'bob35')
        self.assertEqual(PartTestTemplate.generateTestKey('BO B%35'), 'bob35')
        self.assertEqual(PartTestTemplate.generateTestKey('      %  '), '')
        self.assertEqual(PartTestTemplate.generateTestKey(''), '')



# -*- coding: utf-8 -*-

import logging
import os

import inventree.base

logger = logging.getLogger('inventree')


class LabelPrintingMixin:
    """Mixin class for label printing."""

    LABELNAME = ''
    LABELITEM = ''

    def getTemplateId(self, template):
        """Return the ID (pk) from the supplied template."""

        if type(template) in [str, int]:
            return int(template)
        
        if hasattr(template, 'pk'):
            return int(template.pk)
        
        raise ValueError(f"Provided label template is not a valid type: {type(template)}")

    def saveOutput(self, output, filename):
        """Save the output from a label printing job to the specified file path."""

        if os.path.exists(filename) and os.path.isdir(filename):
            filename = os.path.join(
                filename,
                f'Label_{self.getModelType()}_{self.pk}.pdf'
            )
        
        return self._api.downloadFile(url=output, destination=filename)

    def printLabel(self, template, plugin=None, destination=None, *args, **kwargs):
        """Print a label against the provided label template."""

        print_url = '/label/print/'

        template_id = self.getTemplateId(template)

        data = {
            'template': template_id,
            'items': [self.pk]
        }

        if plugin is not None:
            # For the modern printing API, plugin is provided as a key (string) value
            if type(plugin) is str:
                pass
            elif hasattr(plugin, 'key'):
                plugin = plugin.key
            else:
                raise ValueError(f"Invalid plugin provided: {type(plugin)}")
            
            data['plugin'] = plugin
        
        response = self._api.post(
            print_url,
            data=data
        )

        output = response.get('output', None)

        if output and destination:
            return self.saveOutput(output, destination)
        else:
            return response

    def getLabelTemplates(self, **kwargs):
        """Return a list of label templates for this model class."""

        return LabelTemplate.list(
            self._api,
            model_type=self.getModelType(),
            **kwargs
        )


class LabelFunctions(inventree.base.MetadataMixin, inventree.base.InventreeObject):
    """Base class for label functions."""

    @property
    def template_key(self):
        """Return the attribute name for the template file."""

        return 'template'

    @classmethod
    def create(cls, api, data, label, **kwargs):
        """Create a new label by uploading a label template file. Convenience wrapper around base create() method.

        Args:
            data: Dict of data including at least name and description for the template
            label: Either a string (filename) or a file object
        """

        # POST endpoints for creating new reports were added in API version 156
        cls.MIN_API_VERSION = 156

        try:
            # If label is already a readable object, don't convert it
            if label.readable() is False:
                raise ValueError("Label template file must be readable")
        except AttributeError:
            label = open(label)
            if label.readable() is False:
                raise ValueError("Label template file must be readable")

        try:
            response = super().create(api, data=data, files={'template': label}, **kwargs)
        finally:
            if label is not None:
                label.close()
        return response

    def save(self, data=None, label=None, **kwargs):
        """Save label to database. Convenience wrapper around save() method.

        Args:
            data (optional): Dict of data to change for the template.
            label (optional): Either a string (filename) or a file object, to upload a new label template
        """

        if label is not None:
            try:
                # If template is already a readable object, don't convert it
                if label.readable() is False:
                    raise ValueError("Label template file must be readable")
            except AttributeError:
                label = open(label, 'r')
                if label.readable() is False:
                    raise ValueError("Label template file must be readable")

            if 'files' in kwargs:
                files = kwargs.pop('kwargs')
                files[self.template_key] = label
            else:
                files = {self.template_key: label}
        else:
            files = None

        try:
            response = super().save(data=data, files=files)
        finally:
            if label is not None:
                label.close()
        return response

    def downloadTemplate(self, destination, overwrite=False):
        """Download template file for the label to the given destination"""

        # Use downloadFile method to get the file
        return self._api.downloadFile(url=self._data[self.template_key], destination=destination, overwrite=overwrite)


class LabelTemplate(LabelFunctions):
    """Class representing the LabelTemplate database model."""

    URL = 'label/template'

    def __str__(self):
        """String representation of the LabelTemplate instance."""

        return f"LabelTemplate <{self.pk}>: '{self.name}' - ({self.model_type})"



# -*- coding: utf-8 -*-

import os
import sys

sys.path.append(os.path.abspath(os.path.dirname(__file__)))

from test_api import InvenTreeTestCase  # noqa: E402

from inventree.part import InternalPrice, Part  # noqa: E402


class InternalPriceTest(InvenTreeTestCase):
    """ Test that the InternalPrice related objects can be managed via the API """

    def test_fields(self):
        """
        Test field names via OPTIONS request
        """

        field_names = InternalPrice.fieldNames(self.api)

        for field in [
            'part',
            'quantity',
            'price'
        ]:
            self.assertIn(field, field_names)

    def test_internal_price_create(self):
        """
        Tests the ability to create an internal price
        """
        p = Part.create(self.api, {
            'name': 'Test Part',
            'description': 'Test Part',
            'category': 1,
            'revision': 1,
            'active': True,
        })

        self.assertIsNotNone(p)
        self.assertIsNotNone(p.pk)

        ip = InternalPrice.create(self.api, {
            'part': p.pk,
            'quantity': 1,
            'price': '1.00'
        })

        self.assertIsNotNone(ip)



# -*- coding: utf-8 -*-

import os
import sys

sys.path.append(os.path.abspath(os.path.dirname(__file__)))

from test_api import InvenTreeTestCase  # noqa: E402

from inventree.label import LabelTemplate  # noqa: E402
from inventree.part import Part  # noqa: E402
from inventree.plugin import InvenTreePlugin  # noqa: E402


class LabelTemplateTests(InvenTreeTestCase):
    """Unit tests for label templates and printing, using the modern API."""

    def test_label_template_list(self):
        """List available label templates."""

        n = LabelTemplate.count(self.api)
        templates = LabelTemplate.list(self.api)

        self.assertEqual(n, len(templates))
        self.assertGreater(len(templates), 0)

        # Check some expected attributes
        for attr in ['name', 'description', 'enabled', 'model_type', 'template']:
            for template in templates:
                self.assertIn(attr, template)

        for idx, template in enumerate(templates):
            enabled = idx > 0

            if template.enabled != enabled:
                template.save(data={'enabled': idx > 0})

        # Filter by 'enabled' status
        templates = LabelTemplate.list(self.api, enabled=True)
        self.assertGreater(len(templates), 0)
        self.assertLess(len(templates), n)

        # Filter by 'disabled' status
        templates = LabelTemplate.list(self.api, enabled=False)
        self.assertGreater(len(templates), 0)
        self.assertLess(len(templates), n)

    def test_label_print(self):
        """Print a template!"""

        # Find a part to print
        part = Part.list(self.api, limit=1)[0]

        templates = part.getLabelTemplates()
        self.assertGreater(len(templates), 0)

        template = templates[0]

        # Find an available plugin
        plugins = InvenTreePlugin.list(self.api, active=True, mixin='labels')
        self.assertGreater(len(plugins), 0)

        plugin = plugins[0]

        response = part.printLabel(template, plugin=plugin)

        for key in ['created', 'complete', 'output']:
            self.assertIn(key, response)
        
        self.assertEqual(response['complete'], True)
        self.assertIsNotNone(response['output'])



# -*- coding: utf-8 -*-

import inventree.base
import inventree.report


class Build(
    inventree.base.AttachmentMixin,
    inventree.base.StatusMixin,
    inventree.base.MetadataMixin,
    inventree.report.ReportPrintingMixin,
    inventree.base.InventreeObject,
):
    """ Class representing the Build database model """

    URL = 'build'
    MODEL_TYPE = 'build'

    def issue(self):
        """Mark this build as 'issued'."""
        return self._statusupdate(status='issue')
    
    def hold(self):
        """Mark this build as 'on hold'."""
        return self._statusupdate(status='hold')

    def complete(
        self,
        accept_overallocated='reject',
        accept_unallocated=False,
        accept_incomplete=False,
    ):
        """Finish a build order. Takes the following flags:
        - accept_overallocated
        - accept_unallocated
        - accept_incomplete
        """
        return self._statusupdate(
            status='finish',
            data={
                'accept_overallocated': accept_overallocated,
                'accept_unallocated': accept_unallocated,
                'accept_incomplete': accept_incomplete,
            }
        )

    def finish(self, *args, **kwargs):
        """Alias for complete"""
        return self.complete(*args, **kwargs)



# -*- coding: utf-8 -*-

"""
The inventree_api module handles low level requests and authentication
with the InvenTree database server.
"""

import json
import logging
import os
from urllib.parse import urljoin, urlparse

import requests
from requests.auth import HTTPBasicAuth
from requests.exceptions import Timeout

logger = logging.getLogger('inventree')


class InvenTreeAPI(object):
    """
    Basic class for performing Inventree API requests.
    """

    MIN_SUPPORTED_API_VERSION = 206

    @staticmethod
    def getMinApiVersion():
        """
        Return the minimum supported API version
        """

        return InvenTreeAPI.MIN_SUPPORTED_API_VERSION

    def __init__(self, host=None, **kwargs):
        """ Initialize class with initial parameters

        Args:
            base_url - Base URL for the InvenTree server, including port (if required)
                       e.g. "http://inventree.server.com:8000"

        kwargs:
            username - Login username
            password - Login password
            token - Authentication token (if provided, username/password are ignored)
            token-name - Name of the token to use (default = 'inventree-python-client')
            use_token_auth - Use token authentication? (default = True)
            verbose - Print extra debug messages (default = False)
            strict - Enforce strict HTTPS certificate checking (default = True)
            timeout - Set timeout to use (in seconds). Default: 10
            proxies - Definition of proxies as a dict (defaults to an empty dict)

        Login details can be specified using environment variables, rather than being provided as arguments:
            INVENTREE_API_HOST - Host address e.g. "http://inventree.server.com:8000"
            INVENTREE_API_USERNAME - Username
            INVENTREE_API_PASSWORD - Password
            INVENTREE_API_TOKEN - User access token
            INVENTREE_API_TIMEOUT - Timeout value, in seconds
        """

        self.setHostName(host or os.environ.get('INVENTREE_API_HOST', None))

        # Check for environment variables
        self.username = kwargs.get('username', os.environ.get('INVENTREE_API_USERNAME', None))
        self.password = kwargs.get('password', os.environ.get('INVENTREE_API_PASSWORD', None))
        self.token = kwargs.get('token', os.environ.get('INVENTREE_API_TOKEN', None))
        self.token_name = kwargs.get('token_name', os.environ.get('INVENTREE_API_TOKEN_NAME', 'inventree-python-client'))
        self.timeout = kwargs.get('timeout', os.environ.get('INVENTREE_API_TIMEOUT', 10))
        self.proxies = kwargs.get('proxies', dict())
        self.strict = bool(kwargs.get('strict', True))

        self.use_token_auth = kwargs.get('use_token_auth', True)
        self.verbose = kwargs.get('verbose', False)

        self.auth = None
        self.connected = False

        if kwargs.get('connect', True):
            self.connect()

    def setHostName(self, host):
        """Validate that the provided base URL is valid"""

        if host is None:
            raise AttributeError("InvenTreeAPI initialized without providing host address")

        # Ensure that the provided URL is valid
        url = urlparse(host)

        if not url.scheme:
            raise Exception(f"Host '{host}' supplied without valid scheme")

        if not url.netloc or not url.hostname:
            raise Exception(f"Host '{host}' supplied without valid hostname")

        # Check if the path is provided with '/api/' at the end
        ps = [el for el in url.path.split('/') if len(el) > 0]

        if len(ps) > 0 and ps[-1] == 'api':
            ps = ps[:-1]

        path = '/'.join(ps)

        # Re-construct the URL as required
        self.base_url = f"{url.scheme}://{url.netloc}/{path}"

        if not self.base_url.endswith('/'):
            self.base_url += '/'

        # Re-construct the API URL as required
        self.api_url = urljoin(self.base_url, 'api/')

    def connect(self):
        """Attempt a connection to the server"""

        logger.info(f"Connecting to server: {self.base_url}")

        self.connected = False

        # Check if the server is there
        try:
            self.connected = self.testServer()
        except Timeout as e:
            # Send the Timeout error further
            raise e
        except Exception:
            raise ConnectionRefusedError("Could not connect to InvenTree server")

        # Basic authentication
        self.auth = HTTPBasicAuth(self.username, self.password)

        if not self.testAuth():
            raise ConnectionError("Authentication at InvenTree server failed")

        if self.use_token_auth:
            if not self.token:
                self.requestToken()

    def constructApiUrl(self, endpoint_url):
        """Construct an API endpoint URL based on the provided API URL.

        Arguments:
            endpoint_url: The particular API endpoint (everything after "/api/")

        Returns: A fully qualified URL for the subsequent request
        """

        # Strip leading / character if provided
        if endpoint_url.startswith("/"):
            endpoint_url = endpoint_url[1:]

        url = urljoin(self.api_url, endpoint_url)

        # Ensure the API URL ends with a trailing slash
        if not url.endswith('/'):
            url += '/'

        return url

    def testAuth(self):
        """
        Checks if the set user credentials or the used token
        are valid and raises an exception if not.
        """
        logger.info("Checking InvenTree user credentials")

        if not self.connected:
            logger.fatal("InvenTree server is not connected. Skipping authentication check")
            return False

        try:
            response = self.get('/user/me/')
        except requests.exceptions.HTTPError as e:
            logger.fatal(f"Authentication error: {str(type(e))}")
            return False
        except Exception as e:
            logger.fatal(f"Unhandled server error: {str(type(e))}")
            # Re-throw the exception
            raise e

        # set user_name if not initially set
        if not self.username:
            self.username = response['username']

        return True

    def testServer(self):
        """
        Check to see if the server is present.
        The InvenTree server provides a simple endpoint at /api/
        which contains some simple data (and does not require authentication)
        """

        self.server_details = None

        logger.info("Checking InvenTree server connection...")

        try:
            response = requests.get(
                self.api_url,
                timeout=self.timeout,
                proxies=self.proxies,
                verify=self.strict
            )
        except requests.exceptions.ConnectionError as e:
            logger.fatal(f"Server connection error: {str(type(e))}")
            return False
        except Exception as e:
            logger.fatal(f"Unhandled server error: {str(type(e))}")
            # Re-throw the exception
            raise e

        if response.status_code != 200:
            raise requests.exceptions.RequestException(f"Error code from server: {response.status_code} - {response.text}")

        # Record server details
        self.server_details = json.loads(response.text)

        logger.info(f"InvenTree server details: {response.text}")

        # The details provided by the server should include some specific data:
        server_name = str(self.server_details.get('server', ''))

        if not server_name.lower() == 'inventree':
            logger.warning(f"Server returned strange response (expected 'InvenTree', found '{server_name}')")

        api_version = self.server_details.get('apiVersion', '1')

        try:
            api_version = int(api_version)
        except ValueError:
            raise ValueError(f"Server returned invalid API version: '{api_version}'")

        if api_version < InvenTreeAPI.getMinApiVersion():
            raise ValueError(f"Server API version ({api_version}) is older than minimum supported API version ({InvenTreeAPI.getMinApiVersion()})")

        # Store the server API version
        self.api_version = api_version

        return True

    def requestToken(self):
        """ Return authentication token from the server """

        if not self.username or not self.password:
            raise AttributeError('Supply username and password to request token')

        logger.info("Requesting auth token from server...")

        if not self.connected:
            logger.fatal("InvenTree server is not connected. Skipping token request")
            return False

        # Request an auth token from the server
        try:
            response = self.get(
                '/user/token/',
                params={
                    'name': self.token_name,
                }
            )
        except Exception as e:
            logger.error(f"Error requesting token: {str(type(e))}")
            return None

        if 'token' not in response:
            logger.error(f"Token not returned by server: {response}")
            return None

        self.token = response['token']

        logger.info(f"Authentication token: {self.token}")

        return self.token

    def request(self, api_url, **kwargs):
        """ Perform a URL request to the Inventree API """

        if not self.connected:
            # If we have not established a connection to the server yet, attempt now
            self.connect()

        api_url = self.constructApiUrl(api_url)

        data = kwargs.get('data', kwargs.get('json', {}))
        files = kwargs.get('files', {})
        params = kwargs.get('params', {})
        headers = kwargs.get('headers', {})
        proxies = kwargs.get('proxies', self.proxies)

        search_term = kwargs.pop('search', None)

        if search_term is not None:
            params['search'] = search_term

        # Use provided HTTP method
        method = kwargs.get('method', 'get')

        methods = {
            'GET': requests.get,
            'POST': requests.post,
            'PUT': requests.put,
            'PATCH': requests.patch,
            'DELETE': requests.delete,
            'OPTIONS': requests.options,
        }

        if method.upper() not in methods.keys():
            logger.error(f"Unknown request method '{method}'")
            return None

        method = method.upper()

        payload = {
            'params': params,
            'timeout': kwargs.get('timeout', self.timeout),
        }

        if self.use_token_auth and self.token:
            headers['AUTHORIZATION'] = f'Token {self.token}'
            auth = None
        else:
            auth = self.auth

        payload['headers'] = headers
        payload['auth'] = auth
        payload['proxies'] = proxies

        # If we are providing files, we cannot upload as a 'json' request
        if files:
            payload['data'] = data
            payload['files'] = files
        else:
            payload['json'] = data

        # Enforce strict HTTPS certificate checking?
        payload['verify'] = self.strict

        # Debug request information
        logger.debug("Sending Request:")
        logger.debug(f" - URL: {method} {api_url}")

        for item, value in payload.items():
            logger.debug(f" - {item}: {value}")

        # Send request to server!
        try:
            response = methods[method](api_url, **payload)
        except Timeout as e:
            # Re-throw Timeout, and add a message to the log
            logger.critical(f"Server timed out during api.request - {method} @ {api_url}. Timeout {payload['timeout']} s.")
            raise e
        except Exception as e:
            # Re-thrown any caught errors, and add a message to the log
            logger.critical(f"Error at api.request - {method} @ {api_url}")
            raise e

        if response is None:
            logger.error(f"Null response - {method} '{api_url}'")
            return None

        logger.info(f"Request: {method} {api_url} - {response.status_code}")

        # Detect invalid response codes
        # Anything 300+ is 'bad'
        if response.status_code >= 300:

            detail = {
                'detail': 'Error occurred during API request',
                'url': api_url,
                'method': method,
                'status_code': response.status_code,
                'body': response.text,
            }

            if headers:
                detail['headers'] = headers

            if params:
                detail['params'] = params

            if files:
                detail['files'] = files

            if data:
                detail['data'] = data

            raise requests.exceptions.HTTPError(detail)

        # A delete request won't return JSON formatted data (ignore further checks)
        if method == 'DELETE':
            return response

        ctype = response.headers.get('content-type')

        # An API request must respond in JSON format
        if not ctype == 'application/json':
            raise requests.exceptions.InvalidJSONError(
                f"'Response content-type is not JSON - '{api_url}' - '{ctype}'"
            )

        return response

    def delete(self, url, **kwargs):
        """ Perform a DELETE request. Used to remove a record in the database.

        """

        headers = kwargs.get('headers', {})

        response = self.request(url, method='delete', headers=headers, **kwargs)

        if response is None:
            return None

        if response.status_code not in [204]:
            logger.error(f"DELETE request failed at '{url}' - {response.status_code}")

        logger.debug(f"DELETE request at '{url}' returned: {response.status_code} {response.text}")

        return response

    def post(self, url, data, **kwargs):
        """ Perform a POST request. Used to create a new record in the database.

        Args:
            url - API endpoint URL
            data - JSON data to create new object
            files - Dict of file attachments
        """

        params = {
            'format': kwargs.pop('format', 'json')
        }

        response = self.request(
            url,
            json=data,
            method='post',
            params=params,
            **kwargs
        )

        if response is None:
            logger.error(f"PATCH returned null response at '{url}'")
            return None

        if response.status_code not in [200, 201]:
            logger.error(f"PATCH request failed at '{url}' - {response.status_code}")
            return None

        try:
            data = json.loads(response.text)
        except json.decoder.JSONDecodeError:
            logger.error(f"Error decoding JSON response - '{url}'")
            return None

        return data

    def patch(self, url, data, **kwargs):
        """
        Perform a PATCH request.

        Args:
            url - API endpoint URL
            data - JSON data
            files - optional FILES struct
        """

        params = {
            'format': kwargs.pop('format', 'json')
        }

        response = self.request(
            url,
            json=data,
            method='patch',
            params=params,
            **kwargs
        )

        if response is None:
            logger.error(f"PATCH returned null response at '{url}'")
            return None

        if response.status_code not in [200, 201]:
            logger.error(f"PATCH request failed at '{url}' - {response.status_code}")
            return None

        try:
            data = json.loads(response.text)
        except json.decoder.JSONDecodeError:
            logger.error(f"Error decoding JSON response - '{url}'")
            return None

        return data

    def put(self, url, data, **kwargs):
        """
        Perform a PUT request. Used to update existing records in the database.

        Args:
            url - API endpoint URL
            data - JSON data to PUT
        """

        params = {
            'format': kwargs.pop('format', 'json')
        }

        response = self.request(
            url,
            json=data,
            method='put',
            params=params,
            **kwargs
        )

        if response is None:
            return None

        if response.status_code not in [200, 201]:
            logger.error(f"PUT request failed at '{url}' - {response.status_code}")
            return None

        try:
            data = json.loads(response.text)
        except json.decoder.JSONDecodeError:
            logger.error(f"Error decoding JSON response - '{url}'")
            return None

        return data

    def get(self, url, **kwargs):
        """ Perform a GET request.

        For argument information, refer to the 'request' method
        """

        response = self.request(url, method='get', **kwargs)

        # No response returned
        if response is None:
            return None

        try:
            data = json.loads(response.text)
        except json.decoder.JSONDecodeError:
            logger.error(f"Error decoding JSON response - '{url}'")
            return None

        return data

    def downloadFile(self, url, destination, overwrite=False, params=None, proxies=dict()):
        """
        Download a file from the InvenTree server.

        Args:
            destination: Filename (string)

        - If the "destination" is a directory, use the filename of the remote URL
        """

        if url.startswith('/'):
            url = url[1:]

        fullurl = urljoin(self.base_url, url)

        if os.path.exists(destination) and os.path.isdir(destination):

            destination = os.path.join(
                destination,
                os.path.basename(fullurl)
            )

        destination = os.path.abspath(destination)

        if os.path.exists(destination) and not overwrite:
            raise FileExistsError(f"Destination file '{destination}' already exists")

        if self.token:
            headers = {
                'AUTHORIZATION': f"Token {self.token}"
            }
            auth = None
        else:
            headers = {}
            auth = self.auth

        with requests.get(
            fullurl,
            stream=True,
            auth=auth,
            headers=headers,
            params=params,
            timeout=self.timeout,
            proxies=self.proxies,
            verify=self.strict,
        ) as response:

            # Error code
            if response.status_code >= 300:
                detail = {
                    'detail': 'Error occurred during file download',
                    'url': url,
                    'status_code': response.status_code,
                    'body': response.text
                }

                if headers:
                    detail['headers'] = headers

                raise requests.exceptions.HTTPError(detail)

            headers = response.headers

            if not url.startswith('media/report') and not url.startswith('media/label') and 'Content-Type' in headers and 'text/html' in headers['Content-Type']:
                logger.error(f"Error downloading file '{url}': Server return invalid response (text/html)")
                return False

            with open(destination, 'wb') as f:

                for chunk in response.iter_content(chunk_size=16 * 1024):
                    f.write(chunk)

        logger.info(f"Downloaded '{url}' to '{destination}'")
        return True

    def scanBarcode(self, barcode_data):
        """Scan a barcode to see if it matches a known object"""

        if type(barcode_data) is dict:
            barcode_data = json.dumps(barcode_data)

        response = self.post(
            '/barcode/',
            {
                'barcode': str(barcode_data),
            }
        )

        return response



# -*- coding: utf-8 -*-

import inventree.base


class InvenTreePlugin(inventree.base.MetadataMixin, inventree.base.InventreeObject):
    """Represents a PluginConfig instance on the InvenTree server."""

    URL = 'plugins'
    MIN_API_VERSION = 197

    @classmethod
    def getPkField(cls):
        """Return the primary key field for the PluginConfig object."""
        return 'key'

    def setActive(self, active: bool):
        """Activate or deactivate this plugin."""

        url = f'plugins/{self.pk}/activate/'

        self._api.post(url, data={'active': active})



"""
Sales Order models
"""

import inventree.base
import inventree.company
import inventree.part
import inventree.report
import inventree.stock


class SalesOrder(
    inventree.base.AttachmentMixin,
    inventree.base.MetadataMixin,
    inventree.base.StatusMixin,
    inventree.report.ReportPrintingMixin,
    inventree.base.InventreeObject,
):
    """ Class representing the SalesOrder database model """

    URL = 'order/so'
    MODEL_TYPE = 'salesorder'

    def getCustomer(self):
        """Return the customer associated with this order"""
        return inventree.company.Company(self._api, self.customer)

    def getContact(self):
        """Return the contact associated with this order"""
        if self.contact is not None:
            return inventree.company.Contact(self._api, self.contact)
        else:
            return None

    def getLineItems(self, **kwargs):
        """ Return the line items associated with this order """
        return SalesOrderLineItem.list(self._api, order=self.pk, **kwargs)

    def getExtraLineItems(self, **kwargs):
        """ Return the line items associated with this order """
        return SalesOrderExtraLineItem.list(self._api, order=self.pk, **kwargs)

    def addLineItem(self, **kwargs):
        """
        Create (and return) new SalesOrderLineItem object against this SalesOrder
        """

        kwargs['order'] = self.pk

        return SalesOrderLineItem.create(self._api, data=kwargs)

    def addExtraLineItem(self, **kwargs):
        """
        Create (and return) new SalesOrderExtraLineItem object against this SalesOrder
        """

        kwargs['order'] = self.pk

        return SalesOrderExtraLineItem.create(self._api, data=kwargs)

    def getShipments(self, **kwargs):
        """ Return the shipments associated with this order """

        return SalesOrderShipment.list(self._api, order=self.pk, **kwargs)

    def addShipment(self, reference, **kwargs):
        """ Create (and return) new SalesOrderShipment
        against this SalesOrder """

        kwargs['order'] = self.pk
        kwargs['reference'] = reference

        return SalesOrderShipment.create(self._api, data=kwargs)

    def issue(self, **kwargs):
        """Issue (send) this order"""
        return self._statusupdate(status='issue', **kwargs)

    def hold(self, **kwargs):
        """Place this order on hold"""
        return self._statusupdate(status='hold', **kwargs)

    def cancel(self, **kwargs):
        """Cancel this order"""
        return self._statusupdate(status='cancel', **kwargs)


class SalesOrderLineItem(
    inventree.base.InventreeObject,
    inventree.base.MetadataMixin,
):
    """ Class representing the SalesOrderLineItem database model """

    URL = 'order/so-line'

    def getPart(self):
        """
        Return the Part object referenced by this SalesOrderLineItem
        """
        return inventree.part.Part(self._api, self.part)

    def getOrder(self):
        """
        Return the SalesOrder to which this SalesOrderLineItem belongs
        """
        return SalesOrder(self._api, self.order)

    def allocateToShipment(self, shipment, stockitems=None, quantity=None):
        """
        Assign the items of this line to the given shipment.

        By default, assign the total quantity using the first stock
        item(s) found. As many items as possible, up to the quantity in
        sales order, are assigned.

        To limit which stock items can be used, supply a list of stockitems
        to use in the argument stockitems.

        To limit how many items are assigned, supply a quantity to the
        argument quantity. This can also be used to over-assign the items,
        as no check for the amounts in the sales order is performed.

        This function returns a list of items assigned during this call.
        If nothing is returned, this means that nothing was assigned,
        possibly because no stock items are available.
        """

        # If stockitems are not defined, get the list of available stock items
        if stockitems is None:
            stockitems = self.getPart().getStockItems(include_variants=False, in_stock=True, available=True)

        # If no quantity is defined, calculate the number of required items
        # This is the number of sold items not yet allocated, but can not
        # be higher than the number of allocated items
        if quantity is None:
            required_amount = min(
                self.quantity - self.allocated, self.available_stock
            )

        else:
            try:
                required_amount = int(quantity)
            except ValueError:
                raise ValueError(
                    "Argument quantity must be convertible to an integer"
                )

        # Look through stock items, assign items until the required amount
        # is reached
        items = list()

        for SI in stockitems:

            # Check if we are done
            if required_amount <= 0:
                continue

            # Check that this item has available stock
            if SI.quantity - SI.allocated > 0:
                thisitem = {
                    "line_item": self.pk,
                    "quantity": min(
                        required_amount, SI.quantity - SI.allocated
                    ),
                    "stock_item": SI.pk
                }

                # Correct the required amount
                required_amount -= thisitem["quantity"]

                # Append
                items.append(thisitem)

        # Use SalesOrderShipment method to perform allocation
        if len(items) > 0:
            return shipment.allocateItems(items)


class SalesOrderExtraLineItem(
    inventree.base.InventreeObject,
    inventree.base.MetadataMixin,
):
    """ Class representing the SalesOrderExtraLineItem database model """

    URL = 'order/so-extra-line'

    def getOrder(self):
        """
        Return the SalesOrder to which this SalesOrderLineItem belongs
        """
        return SalesOrder(self._api, self.order)


class SalesOrderAllocation(
    inventree.base.InventreeObject
):
    """Class representing the SalesOrderAllocation database model."""

    MIN_API_VERSION = 267
    URL = 'order/so-allocation'

    def getOrder(self):
        """Return the SalesOrder to which this SalesOrderAllocation belongs."""
        return SalesOrder(self._api, self.order)

    def getShipment(self):
        """Return the SalesOrderShipment to which this SalesOrderAllocation belongs."""
        from sales_order import SalesOrderShipment
        return SalesOrderShipment(self._api, self.shipment)

    def getLineItem(self):
        """Return the SalesOrderLineItem to which this SalesOrderAllocation belongs."""
        return SalesOrderLineItem(self._api, self.line)

    def getStockItem(self):
        """Return the StockItem to which this SalesOrderAllocation belongs."""
        return inventree.stock.StockItem(self._api, self.item)

    def getPart(self):
        """Return the Part to which this SalesOrderAllocation belongs."""
        return inventree.part.Part(self._api, self.part)


class SalesOrderShipment(
    inventree.base.InventreeObject,
    inventree.base.StatusMixin,
    inventree.base.MetadataMixin,
):
    """Class representing a shipment for a SalesOrder"""

    URL = 'order/so/shipment'

    def getOrder(self):
        """Return the SalesOrder to which this SalesOrderShipment belongs."""
        return SalesOrder(self._api, self.order)

    def allocateItems(self, items=[]):
        """
        Function to allocate items to the current shipment

        items is expected to be a list containing dicts, one for each item
        to be assigned. Each dict should contain three parameters, as
        follows:
            items = [{
                "line_item": 25,
                "quantity": 150,
                "stock_item": 26
            }
        """

        # Customize URL
        url = f'order/so/{self.getOrder().pk}/allocate'

        # Create data from given inputs
        data = {
            'items': items,
            'shipment': self.pk
        }

        # Send data
        response = self._api.post(url, data)

        # Reload
        self.reload()

        # Return
        return response

    def getAllocations(self):
        """Return the allocations associated with this shipment"""
        return SalesOrderAllocation.list(self._api, shipment=self.pk)

    @property
    def allocations(self):
        """Return the allocations associated with this shipment.

        Note: This is an overload of getAllocations() method, for legacy compatibility.
        """
        try:
            return self.getAllocations()
        except NotImplementedError:
            return self._data['allocations']

    def complete(
        self,
        shipment_date=None,
        tracking_number='',
        invoice_number='',
        link='',
        **kwargs
    ):
        """
        Complete the shipment, with given shipment_date, or reasonable
        defaults.
        """

        # Create data from given inputs
        data = {
            'shipment_date': shipment_date,
            'tracking_number': tracking_number,
            'invoice_number': invoice_number,
            'link': link
        }

        # Return
        return self._statusupdate(status='ship', data=data, **kwargs)

    def ship(self, *args, **kwargs):
        """Alias for complete function"""
        self.complete(*args, **kwargs)



"""Unit tests for the ProjectCode model"""

import os
import sys

sys.path.append(os.path.abspath(os.path.dirname(__file__)))

from test_api import InvenTreeTestCase  # noqa: E402

from inventree.project_code import ProjectCode  # noqa: E402


class ProjectCodeTest(InvenTreeTestCase):
    """Tests for the ProjectCode model."""

    def test_project_code_create(self):
        """Test we can create a new project code."""

        n = ProjectCode.count(self.api)

        ProjectCode.create(self.api, {
            'code': f'TEST {n + 1}',
            'description': 'Test project code',
        })

        self.assertEqual(ProjectCode.count(self.api), n + 1)

        # Try to create a duplicate code
        with self.assertRaises(Exception):
            ProjectCode.create(self.api, {
                'code': f'TEST {n + 1}',
                'description': 'Test project code',
            })

        n = ProjectCode.count(self.api)

        # Create 5 more codes
        for idx in range(5):
            ProjectCode.create(self.api, {
                'code': f'CODE-{idx + n}',
                'description': f'Description {idx + n}',
            })

        # List all codes
        codes = ProjectCode.list(self.api)

        self.assertEqual(len(codes), n + 5)




import logging

import inventree.base

logger = logging.getLogger('inventree')


class ProjectCode(inventree.base.InventreeObject):
    """Class representing the 'ProjectCode' database model"""

    URL = 'project-code/'
    MIN_API_VERSION = 109



# -*- coding: utf-8 -*-

import logging

import inventree.base

logger = logging.getLogger('inventree')


class User(inventree.base.InventreeObject):
    """ Class representing the User database model """

    URL = 'user'



# -*- coding: utf-8 -*-

import logging
import re

import inventree.base
import inventree.build
import inventree.company
import inventree.label
import inventree.report
import inventree.stock

logger = logging.getLogger('inventree')


class PartCategoryParameterTemplate(inventree.base.InventreeObject):
    """A model which link a ParameterTemplate to a PartCategory"""

    URL = 'part/category/parameters'

    def getCategory(self):
        """Return the referenced PartCategory instance"""
        return PartCategory(self._api, self.category)

    def getTemplate(self):
        """Return the referenced ParameterTemplate instance"""
        return ParameterTemplate(self._api, self.parameter_template)


class PartCategory(inventree.base.MetadataMixin, inventree.base.InventreeObject):
    """ Class representing the PartCategory database model """

    URL = 'part/category'

    def getParts(self, **kwargs):
        return Part.list(self._api, category=self.pk, **kwargs)

    def getParentCategory(self):
        if self.parent:
            return PartCategory(self._api, self.parent)
        else:
            return None

    def getChildCategories(self, **kwargs):
        return PartCategory.list(self._api, parent=self.pk, **kwargs)

    def getCategoryParameterTemplates(self, fetch_parent: bool = True) -> list:
        """Fetch a list of default parameter templates associated with this category

        Arguments:
            fetch_parent: If True (default) include templates for parents also
        """

        return PartCategoryParameterTemplate.list(
            self._api,
            category=self.pk,
            fetch_parent=fetch_parent
        )


class Part(
    inventree.base.AttachmentMixin,
    inventree.base.BarcodeMixin,
    inventree.base.MetadataMixin,
    inventree.base.ImageMixin,
    inventree.label.LabelPrintingMixin,
    inventree.base.InventreeObject,
):
    """ Class representing the Part database model """

    URL = 'part'
    MODEL_TYPE = 'part'

    def getCategory(self):
        """ Return the part category associated with this part """
        return PartCategory(self._api, self.category)

    def getTestTemplates(self):
        """ Return all test templates associated with this part """
        return PartTestTemplate.list(self._api, part=self.pk)

    def getSupplierParts(self):
        """ Return the supplier parts associated with this part """
        if self.purchaseable:
            return inventree.company.SupplierPart.list(self._api, part=self.pk)
        else:
            return list()

    def getManufacturerParts(self):
        """ Return the manufacturer parts associated with this part """
        return inventree.company.ManufacturerPart.list(self._api, part=self.pk)

    def getBomItems(self, **kwargs):
        """ Return the items required to make this part """
        return BomItem.list(self._api, part=self.pk, **kwargs)

    def isUsedIn(self):
        """ Return a list of all the parts this part is used in """
        return BomItem.list(self._api, uses=self.pk)

    def getBuilds(self, **kwargs):
        """ Return the builds associated with this part """
        return inventree.build.Build.list(self._api, part=self.pk, **kwargs)

    def getStockItems(self, **kwargs):
        """ Return the stock items associated with this part """
        return inventree.stock.StockItem.list(self._api, part=self.pk, **kwargs)

    def getParameters(self):
        """ Return parameters associated with this part """
        return Parameter.list(self._api, part=self.pk)

    def getRelated(self):
        """ Return related parts associated with this part """
        return PartRelated.list(self._api, part=self.pk)

    def getInternalPriceList(self):
        """
        Returns the InternalPrice list for this part
        """

        return InternalPrice.list(self._api, part=self.pk)

    def setInternalPrice(self, quantity: int, price: float):
        """
        Set the internal price for this part
        """

        return InternalPrice.setInternalPrice(self._api, self.pk, quantity, price)
    
    def getSalePrice(self):
        """
        Get sales prices for this part
        """
        return SalePrice.list(self._api, part=self.pk)[0].price

    def getRequirements(self):
        """
        Get required amounts from requirements API endpoint for this part
        """

        # Set the url
        URL = f"{self.URL}/{self.pk}/requirements/"

        # Get data
        return self._api.get(URL)


class PartTestTemplate(inventree.base.MetadataMixin, inventree.base.InventreeObject):
    """ Class representing a test template for a Part """

    URL = 'part/test-template'

    @classmethod
    def generateTestKey(cls, test_name):
        """ Generate a 'key' for this test """

        key = test_name.strip().lower()
        key = key.replace(' ', '')

        # Remove any characters that cannot be used to represent a variable
        key = re.sub(r'[^a-zA-Z0-9]', '', key)

        return key

    def getTestKey(self):
        """Return the 'key' for this test.

        Note that after API v169, the 'key' parameter is also directly accessible
        """

        # Try to return the key - fall back to generateTestKey
        try:
            return self.key
        except AttributeError:
            return self.generateTestKey(self.test_name)


class BomItem(
    inventree.base.InventreeObject,
    inventree.base.MetadataMixin,
    inventree.report.ReportPrintingMixin,
):
    """ Class representing the BomItem database model """

    URL = 'bom'


class InternalPrice(inventree.base.InventreeObject):
    """ Class representing the InternalPrice model """

    URL = 'part/internal-price'

    @classmethod
    def setInternalPrice(cls, api, part, quantity: int, price: float):
        """
        Set the internal price for this part
        """

        data = {
            'part': part,
            'quantity': quantity,
            'price': price,
        }

        # Send the data to the server
        return api.post(cls.URL, data)


class SalePrice(inventree.base.InventreeObject):
    """ Class representing the SalePrice model """

    URL = 'part/sale-price'

    @classmethod
    def setSalePrice(cls, api, part, quantity: int, price: float, price_currency: str):
        """
        Set the sale price for this part
        """

        data = {
            'part': part,
            'quantity': quantity,
            'price': price,
            'price_currency': price_currency,
        }

        # Send the data to the server
        return api.post(cls.URL, data)


class PartRelated(inventree.base.InventreeObject):
    """ Class representing a relationship between parts"""

    URL = 'part/related'

    @classmethod
    def add_related(cls, api, part1, part2):

        if isinstance(part1, Part):
            pk_1 = part1.pk
        else:
            pk_1 = int(part1)
        if isinstance(part2, Part):
            pk_2 = part2.pk
        else:
            pk_2 = int(part2)

        data = {
            'part_1': pk_1,
            'part_2': pk_2,
        }

        # Send the data to the server
        return api.post(cls.URL, data)


class Parameter(inventree.base.InventreeObject):
    """class representing the Parameter database model """
    URL = 'part/parameter'

    def getunits(self):
        """ Get the units for this parameter """

        return self._data['template_detail']['units']


class ParameterTemplate(inventree.base.InventreeObject):
    """ class representing the Parameter Template database model"""

    URL = 'part/parameter/template'



# -*- coding: utf-8 -*-

import logging
import os

import inventree.api
import inventree.base
import inventree.company
import inventree.label
import inventree.part
import inventree.report


class StockLocation(
    inventree.base.BarcodeMixin,
    inventree.base.MetadataMixin,
    inventree.label.LabelPrintingMixin,
    inventree.report.ReportPrintingMixin,
    inventree.base.InventreeObject,
):
    """ Class representing the StockLocation database model """

    URL = 'stock/location'
    MODEL_TYPE = 'stocklocation'

    def getStockItems(self, **kwargs):
        return StockItem.list(self._api, location=self.pk, **kwargs)

    def getParentLocation(self):
        """
        Return the parent stock location
        (or None if no parent is available)
        """

        if self.parent is None:
            return None

        return StockLocation(self._api, pk=self.parent)

    def getChildLocations(self, **kwargs):
        """
        Return all the child locations under this location
        """
        return StockLocation.list(self._api, parent=self.pk, **kwargs)


class StockItem(
    inventree.base.AttachmentMixin,
    inventree.base.BarcodeMixin,
    inventree.base.BulkDeleteMixin,
    inventree.base.MetadataMixin,
    inventree.label.LabelPrintingMixin,
    inventree.base.InventreeObject,
):
    """Class representing the StockItem database model."""

    URL = 'stock'

    MODEL_TYPE = 'stockitem'

    @classmethod
    def adjustStockItems(cls, api: inventree.api.InvenTreeAPI, method: str, items: list, **kwargs):
        """Perform a generic stock 'adjustment' action.

        Arguments:
            api: InvenTreeAPI instance
            method: Adjument method, e.g. 'count' / 'add'
            items: List of items to include in the adjustment (see below)
            kwargs: Additional kwargs to send with the adjustment

        Items:
            Each 'item' in the 'items' list must be a dict object, containing the following fields:

            pk: The 'pk' (primary key) identifier for a StockItem instance
            quantity: The quantity of each stock item for the particular action
        """

        if method not in ['count', 'add', 'remove', 'transfer', 'assign']:
            raise ValueError(f"Stock adjustment method '{method}' not supported")

        url = f"stock/{method}/"

        data = kwargs
        data['items'] = items

        return api.post(url, data=data)

    @classmethod
    def countStockItems(cls, api: inventree.api.InvenTreeAPI, items: list, **kwargs):
        """Perform 'count' adjustment for multiple stock items"""

        return cls.adjustStockItems(
            api,
            'count',
            items,
            **kwargs
        )

    @classmethod
    def addStockItems(cls, api: inventree.api.InvenTreeAPI, items: list, **kwargs):
        """Perform 'add' adjustment for multiple stock items"""

        return cls.adjustStockItems(
            api,
            'add',
            items,
            **kwargs
        )

    @classmethod
    def removeStockItems(cls, api: inventree.api.InvenTreeAPI, items: list, **kwargs):
        """Perform 'remove' adjustment for multiple stock items"""

        return cls.adjustStockItems(
            api,
            'remove',
            items,
            **kwargs
        )

    @classmethod
    def transferStockItems(cls, api: inventree.api.InvenTreeAPI, items: list, location: int, **kwargs):
        """Perform 'transfer' adjustment for multiple stock items"""

        kwargs['location'] = location

        return cls.adjustStockItems(
            api,
            'transfer',
            items,
            **kwargs
        )

    @classmethod
    def assignStockItems(cls, api: inventree.api.InvenTreeAPI, items: list, customer: int, **kwargs):
        """Perform 'assign' adjustment for multiple stock items"""

        kwargs['customer'] = customer

        return cls.adjustStockItems(
            api,
            'assign',
            items,
            **kwargs
        )

    def countStock(self, quantity, **kwargs):
        """Perform a count (stocktake) action for this StockItem"""

        self.countStockItems(
            self._api,
            [
                {
                    'pk': self.pk,
                    'quantity': quantity,
                }
            ],
            **kwargs
        )

    def addStock(self, quantity, **kwargs):
        """Manually add the specified quantity to this StockItem"""

        self.addStockItems(
            self._api,
            [
                {
                    'pk': self.pk,
                    'quantity': quantity,
                }
            ],
            **kwargs
        )

    def removeStock(self, quantity, **kwargs):
        """Manually remove the specified quantity to this StockItem"""

        self.removeStockItems(
            self._api,
            [
                {
                    'pk': self.pk,
                    'quantity': quantity,
                }
            ],
            **kwargs
        )

    def transferStock(self, location, quantity=None, **kwargs):
        """Transfer this StockItem into the specified location.

        Arguments:
            location: A StockLocation instance or integer ID value
            quantity: Optionally specify quantity to transfer. If None, entire quantity is transferred
            notes: Optional transaction notes
        """

        if isinstance(location, StockLocation):
            location = location.pk

        if quantity is None:
            quantity = self.quantity

        self.transferStockItems(
            self._api,
            [
                {
                    'pk': self.pk,
                    'quantity': quantity,
                }
            ],
            location=location,
            **kwargs
        )

    def assignStock(self, customer, **kwargs):
        """Assign this stock item to a customer (by company PK)

        Arguments:
            customer: A Company instance or integer ID value
            notes: Optional transaction notes"""

        if isinstance(customer, inventree.company.Company):
            customer = customer.pk

        self.assignStockItems(
            self._api,
            [
                {
                    'item': self.pk,  # In assign API, item is used instead of item
                }
            ],
            customer=customer,
            **kwargs
        )

    def installStock(self, item, **kwargs):
        """Install the given item into this stock item

        Arguments:
            stockItem: A stockItem instance or integer ID value

        kwargs:
            quantity: quantity of installed items
            notes: Optional transaction notes"""

        if isinstance(item, StockItem):
            quantity = kwargs.get('quantity', item.quantity)
            item = item.pk
        else:
            quantity = kwargs.get('quantity', 1)

        kwargs['quantity'] = kwargs.get('quantity', quantity)
        kwargs['stock_item'] = item

        url = f"stock/{self.pk}/install/"

        return self._api.post(url, data=kwargs)

    def uninstallStock(self, location, quantity=1, **kwargs):
        """Uninstalls this item from any stock item

        Arguments:
            location: A StockLocation instance or integer ID value
            quantity: quantity of removed items. defaults to 1.
            notes: Optional transaction notes"""

        if isinstance(location, StockLocation):
            location = location.pk
        kwargs['quantity'] = quantity
        kwargs['stock_item'] = self.pk
        kwargs['location'] = location

        url = f"stock/{self.pk}/uninstall/"

        return self._api.post(url, data=kwargs)

    def getPart(self):
        """ Return the base Part object associated with this StockItem """
        return inventree.part.Part(self._api, self.part)

    def getLocation(self):
        """
        Return the StockLocation associated with this StockItem

        Returns None if there is no linked StockItem
        """

        if self.location is None:
            return None

        return StockLocation(self._api, self.location)

    def getTrackingEntries(self, **kwargs):
        """Return list of StockItemTracking instances associated with this StockItem"""

        return StockItemTracking.list(self._api, item=self.pk, **kwargs)

    def getTestResults(self, **kwargs):
        """ Return all the test results associated with this StockItem """

        return StockItemTestResult.list(self._api, stock_item=self.pk, **kwargs)

    def uploadTestResult(self, test_name, test_result, **kwargs):
        """ Upload a test result against this StockItem """

        return StockItemTestResult.upload_result(self._api, self.pk, test_name, test_result, **kwargs)


class StockItemTracking(inventree.base.InventreeObject):
    """Class representing a StockItem tracking object."""

    URL = 'stock/track'


class StockItemTestResult(
    inventree.base.BulkDeleteMixin,
    inventree.base.MetadataMixin,
    inventree.report.ReportPrintingMixin,
    inventree.base.InventreeObject,
):
    """Class representing a StockItemTestResult object.

    Note: From API version 169 and onwards, the StockItemTestResult object
    must reference a PartTestTemplate object, rather than a test name.

    However, for backwards compatibility, test results can be uploaded using the test name,
    and will be associated with the correct PartTestTemplate on the server.
    """

    URL = 'stock/test'
    MODEL_TYPE = 'stockitem'

    def getTestTemplate(self):
        '''Return the PartTestTemplate item for this StockItemTestResult'''

        return inventree.part.PartTestTemplate(self._api, self.template)

    def getTestKey(self):

        # Get the test key from the PartTestTemplate
        return self.getTestTemplate().getTestKey()

    @classmethod
    def upload_result(cls, api, stock_item, test, result, **kwargs):
        """
        Upload a test result.

        args:
            api: Authenticated InvenTree API object
            stock_item: pk of the StockItem object to upload the test result against
            test: Name of the test (string) or ID of the test template (integer)
            result: Test result (boolean)

        kwargs:
            attachment: Optionally attach a file to the test result
            notes: Add extra notes
            value: Add a "value" to the test (e.g. an actual measurement made during the test)
        """

        attachment = kwargs.get('attachment', None)

        files = {}

        fo = None

        if attachment:
            if os.path.exists(attachment):
                f = os.path.basename(attachment)
                fo = open(attachment, 'rb')
                files['attachment'] = (f, fo)
            else:
                logging.error(f"File does not exist: '{attachment}'")

        notes = kwargs.get('notes', '')
        value = kwargs.get('value', '')

        data = {
            'stock_item': stock_item,
            'result': result,
            'notes': notes,
            'value': value,
        }

        # Determine the method by which the particular test is designated
        # It is either the test "name", or the test "template" ID
        if type(test) is int:
            data['template'] = test
        elif isinstance(test, inventree.part.PartTestTemplate):
            data['template'] = test.pk
        else:
            data['test'] = str(test)

        # Send the data to the server
        if api.post(cls.URL, data, files=files):
            logging.info(f"Uploaded test result: '{test}'")
            ret = True
        else:
            logging.warning("Test upload failed")
            ret = False

        # Ensure the file attachment is closed after use
        if fo:
            fo.close()

        return ret



# -*- coding: utf-8 -*-

import setuptools

from inventree.base import INVENTREE_PYTHON_VERSION

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()


setuptools.setup(
    name="inventree",
    version=INVENTREE_PYTHON_VERSION,
    author="Oliver Walters",
    author_email="oliver.henry.walters@gmail.com",
    description="Python interface for InvenTree inventory management system",
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords="bom, bill of materials, stock, inventory, management, barcode",
    url="https://github.com/inventree/inventree-python/",
    license="MIT",
    packages=setuptools.find_packages(
        exclude=[
            'ci',
            'scripts',
            'test',
        ]
    ),
    install_requires=[
        "requests>=2.27.0",
        "pip-system-certs>=4.0",
    ],
    setup_requires=[
        "wheel",
        "twine",
        "wrapt"
    ],
    python_requires=">=3.9"
)



"""Manages currency / conversion support for InvenTree"""

import logging

logger = logging.getLogger('inventree')


class CurrencyManager(object):
    """Class for managing InvenTree currency support"""

    # Currency API endpoint
    CURRENCY_ENDPOINT = 'currency/exchange/'

    def __init__(self, api):
        """Construct a CurrencyManager instance"""

        # Store internal reference to the API
        self.api = api

        self.base_currency = None
        self.exchange_rates = None

    def refreshExchangeRates(self):
        """Request the server update exchange rates from external service"""

        return self.api.post('currency/refresh/', {})

    def updateFromServer(self):
        """Retrieve currency data from the server"""

        response = self.api.get(self.CURRENCY_ENDPOINT)

        if response is None:
            logger.error("Could not retrieve currency data from InvenTree server")
            return

        self.base_currency = response.get('base_currency', None)
        self.exchange_rates = response.get('exchange_rates', None)

        if self.base_currency is None:
            logger.warning("'base_currency' missing from server response")

        if self.exchange_rates is None:
            logger.warning("'exchange_rates' missing from server response")

    def getBaseCurrency(self, cache=True):
        """Return the base currency code (e.g. 'USD') from the server"""

        if not cache or not self.base_currency:
            self.updateFromServer()

        return self.base_currency

    def getExchangeRates(self, cache=True):
        """Return the exchange rate information from the server"""

        if not cache or not self.exchange_rates:
            self.updateFromServer()

        return self.exchange_rates

    def convertCurrency(self, value, source_currency, target_currency, cache=True):
        """Convert between currencies

        Arguments:
            value: The numerical currency value to be converted
            source_currency: The source currency code (e.g. 'USD')
            target_currency: The target currency code (e.g. 'NZD')
        """

        # Shortcut if the currencies are the same
        if source_currency == target_currency:
            return value

        base = self.getBaseCurrency(cache=cache)
        rates = self.getExchangeRates(cache=cache)

        if base is None:
            raise AttributeError("Base currency information is not available")

        if rates is None:
            raise AttributeError("Exchange rate information is not available")

        if source_currency not in rates:
            raise NameError(f"Source currency code '{source_currency}' not found in exchange rate data")

        if target_currency not in rates:
            raise NameError(f"Target currency code '{target_currency}' not found in exchange rate data")

        return value / rates[source_currency] * rates[target_currency]
